/*
 *  fractal.cc
 * 
 *  Fractal Generator
 * 
 *  Copyright (c) 1998 Brian Martin
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "fractal.h"
#include "fractal_state.h"
#include "defs.h"
#include <time.h>
#include <math.h>

extern char * yy_scan_string(const char *);

#ifdef USE_THREADS

void *processes_requests(void * fract) {

    ((fractal*)fract)->processes_requests();
    pthread_exit(0);		// probably not necessary
}

void fractal::processes_requests() {
    while(1) {
	switch(requests.remove()) { // queue is synchronized
 	case FRACT_RECOMPUTE_REQ:
	    printf("recompute_req\n");
 	    recompute(false);
	    printf("recompute done\n");
 	    break;

	case FRACT_RECOMPUTE_SIZECHANGE_REQ:
	    printf("recompute sizechange\n");
	    recompute(true);
	    break;

 	case FRACT_REFRESH_REQ:
//	    sem_wait(&x_mutex);
//	    XLockDisplay(dpy);
	    printf("processing refresh\n");
	    requests.clear_pending(FRACT_REFRESH_REQ);	
	    refresh_impl();
	    printf("refresh done\n");
//	    XUnlockDisplay(dpy);
//	    sem_post(&x_mutex);
 	    break;
 	}
    }
}

#endif // USE_THREADS

fractal::fractal(fractal_State * _state) : 
    state(_state),
#ifdef USE_THREADS    
    requests(200),
#endif
    hash(NULL)
{ 
    props = NULL;
    hash = NULL;
    position = NULL;
    
    alloc();

    max_val = 40;
    max_iter = 100;

#ifdef USE_THREADS
    num_threads = 1;		// should probably parameterize this sometime..

    table = new eq_table*[num_threads];
    table[0] = new eq_table(state->gui.expr_tree, max_iter, max_val);
    for(int i = 1 ; i < num_threads ; i++) 
      //	table[i] = new eq_table(table[0]);
      table[i] = new eq_table(state->gui.expr_tree, max_iter, max_val);
#else
    table = new eq_table(state->gui.expr_tree, max_iter, max_val);
#endif
    
    glwin = new GLPort(300, 300, state->filename);
    glwin->show();

    dpy = glwin->getDpy();
    drawable = glwin->getWin();

    glEnable(GL_VERTEX_ARRAY);
    glDisable(GL_NORMALIZE);	// we send unit normals for efficiency
				// or at least is seems like we should
				// but the lighting looks real bad when we do.
    glDepthFunc(GL_LESS);	// enable zbuffer
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_COLOR_MATERIAL);
    glShadeModel(GL_SMOOTH);	// gouraud shading

//    puts(glGetString(GL_EXTENSIONS));

#ifdef USE_THREADS    
    assert(!sem_init(&diverg_mutex, 0, 1));
    assert(!sem_init(&x_mutex, 0, 1));

    pthread_attr_t attrs;
    pthread_attr_init(&attrs);
    pthread_attr_setdetachstate(&attrs, PTHREAD_CREATE_DETACHED);

    assert(!pthread_create(&req_thread, &attrs, &::processes_requests, (void *)this));

#endif USE_THREADS
    
    connect_to_method(refresh, this, &fractal::queue_refresh);

    request(FRACT_RECOMPUTE_REQ);
    request(FRACT_REFRESH_REQ);
    request(FRACT_REFRESH_REQ);
}

void fractal::free_position() {

    if(first_pt)
	delete first_pt;
    first_pt = last_pt = NULL;

    for(int i = 0 ; i < state->gui.size.x.value() ; i++) {
	for( int j = 0 ; j < state->gui.size.y.value() ; j++ ) 
	    delete [] position[i][j];
	delete [] position[i];
    }
    delete [] position;
}

void fractal::alloc() {

    if(position)
	free_position();

    if(props)
	delete props;

//     if(hash)
// 	delete hash;

    if(state->gui.size.changed()) 
	state->gui.size.grab();

    int x = (int)state->gui.size.x.value();
    int y = (int)state->gui.size.y.value();
    int z = (int)state->gui.size.z.value();

    //    light.set_radius(state->gui.size.x.value());

    // now we need to determine a good hash size for this monster..
    
//    int curr_try = x*y*z/4+1;	// pulled this outa my ass
    
				// from knuth vol 2 p 349
//    while(((int)(pow(3, curr_try - 1)))%curr_try != 1)
//	curr_try++;
    
//     hash = new Hash(curr_try);

    props = new bitvect(x, y, z, 4);


//     progress.set_text("allocating memory..");
    //    progress.show();
    position = new fractalPos **[x];
    for(int i = 0 ; i < x ; i++) {
// 	progress.update(((double)i)/(state->gui.size.x.value()-1));
	position[i] = new fractalPos *[y];
	for (int j = 0 ; j < y ; j ++) 
	    position[i][j] = new fractalPos[z];
    }
//    progress.hide();
    
}



void fractal::queue_refresh() { 
#ifdef USE_THREADS    
    request(FRACT_REFRESH_REQ);
#else
    refresh_impl();
#endif
}

bool fractal::divergent(double s_val, double a_val, 
			double b_val, double c_val,
			eq_table *my_table, float * color) {

    double initval = s_val*s_val+a_val*a_val+b_val*b_val+c_val*c_val;

    int iters;

    double val = my_table->iter(s_val, a_val, b_val, c_val, iters);

    if(color)			// only set color for middle points
	*color = (float)(initval/val);

//  return val > max_val;
    return iters != max_iter;
}

// gets and traverses neibhor list looking for divergent neibhors  
// (only called on convergent cells)

bool fractal::divergentNeibhor(int a, int b, int c) {

    coordList * neibhors = enumerateNeibhors(a, b, c, 1);
    coordList * neibhorptr = neibhors;
    bool diverger = false;

    while(neibhorptr != NULL) {
	if(props->get(neibhorptr->a,neibhorptr->b,neibhorptr->c,DIVERGES_1)) {
	    diverger = true;
	    break;
	}
	neibhorptr = neibhorptr->next();
    }
    delete neibhors;
    return diverger;
}

#ifdef USE_THREADS    

void * test_divergence(void * fract) { // need this global func for threads

    //    printf("in thread!!\n");
    printf("in thread, fract = %p\n", fract);
    ((fractal *)fract)->test_divergence();
    pthread_exit(0);
}

/*
  this function divides up the space by the X index and serves chunks
  to threads running fractal::test_divergence in a safe manner.
 */
int fractal::get_next_divergence_block() {

    int retval;
    
    sem_wait(&diverg_mutex);
    if(currblock != state->gui.size.x.value())
	retval = currblock++;
    else
	retval = -1;
    sem_post(&diverg_mutex);
    return retval;
}

int fractal::get_thread_num() {

    int retval;
    sem_wait(&diverg_mutex);
    retval = currthread++;
    sem_post(&diverg_mutex);
    return(retval);
}

void fractal::test_divergence() {

    eq_table * my_table = table[get_thread_num()];
    
    int block;

    double s_val;
    double a_val;
    double b_val;

    printf("props = %p\n", props->data);

    while((block = get_next_divergence_block()) != -1) {
	s_val = state->gui.minval.x.value()+block*s_mod;
	a_val = state->gui.minval.y.value();
	for(int j = 0 ; j < state->gui.size.y.value() ; j++) {
	    b_val = state->gui.minval.z.value();
	    for(int k = 0 ; k < state->gui.size.z.value() ; k++) {
		props->set(block,j,k,0,divergent(s_val,
						 a_val,
						 b_val,
						 w_val[0],my_table,0));
		props->set(block,j,k,1,divergent(s_val,
						 a_val,
						 b_val,
						 w_val[1],
						 my_table,
						 &position[block][j][k].color));
		props->set(block,j,k,2,divergent(s_val,
						 a_val,
						 b_val,
						 w_val[2],my_table,0));
		b_val += b_mod;
	    }
	    a_val += a_mod;
	}
    }
}

#endif // USE_THREADS

void fractal::identifySurface(bool sizechange) {

    // first determine divergent points

#ifdef USE_THREADS    
    int n = num_threads;
    pthread_t thread[n];
#endif

    if(sizechange) {
// 	printf("changing size\n");
// 	printf("fuck\n");
	alloc();
	printf("realloc because of size\n");
	state->gui.size.x.entry->setVal(floor(state->gui.size.x.gui_value()));
	state->gui.size.y.entry->setVal(floor(state->gui.size.y.gui_value()));
	state->gui.size.z.entry->setVal(floor(state->gui.size.z.gui_value()));
	state->gui.size.grab();	// get new values
// 	printf("size is %.2f %.2f %.2f\n", state->gui.size.x.value(),
// 	       state->gui.size.y.value(), state->gui.size.z.value());
	//	printf("duck\n");
//	w_val = w_fix*(((double)new_size)/old_size);
	clear();
    }
    
#ifdef USE_THREADS
    
    if(state->gui.q_is_c_initially) 
	for(int i = 0 ; i < n ; i++) 
	    table[i]->set_q_to_c_initially();
    else
	for(int i = 0 ; i < n ; i++) 
	    table[i]->set_init_q_val(state->gui.initial_q_val);	
#else

    if(state->gui.q_is_c_initially) 
	table->set_q_to_c_initially();
    else 
	table->set_init_q_val(state->gui.initial_q_val);

#endif

    double temp = ((double)state->gui.maxval.z.value()-
		   state->gui.minval.z.value())/state->gui.size.z.value();
    w_val[0] = state->gui.w_fix.value()-temp;
    w_val[1] = state->gui.w_fix.value();
    w_val[2] = state->gui.w_fix.value()+temp;

    double s_val;
    double a_val;
    double b_val;
    s_mod = ((double)state->gui.maxval.x.value()-state->gui.minval.x.value())
	/state->gui.size.x.value();
    a_mod = ((double)state->gui.maxval.y.value()-state->gui.minval.y.value())
	/state->gui.size.y.value();
    b_mod = ((double)state->gui.maxval.z.value()-state->gui.minval.z.value())
	/state->gui.size.z.value();

//    progress.set_text("determining divergence..");
    
#ifdef USE_THREADS
    currblock = 0;
    currthread = 0;

    for(int i = 0 ; i < n ; i++)
 	pthread_create(&thread[i], NULL, &::test_divergence, (void *)this);

    for(int i = 0 ; i < n ; i++)
 	pthread_join(thread[i], NULL);
#else

    s_val = state->gui.minval.x.value();
    for(int i = 0 ; i < state->gui.size.x.value() ; i++) {
	a_val = state->gui.minval.y.value();
	for(int j = 0 ; j < state->gui.size.y.value() ; j++) {
	    b_val = state->gui.minval.z.value();
	    for(int k = 0 ; k < state->gui.size.z.value() ; k++) {
		props->set(i,j,k,0,divergent(s_val,
					     a_val,
					     b_val,
					     w_val[0],table,0));
		props->set(i,j,k,1,divergent(s_val,
					     a_val,
					     b_val,
					     w_val[1],
					     table,&position[i][j][k].color));
		props->set(i,j,k,2,divergent(s_val,
					     a_val,
					     b_val,
					     w_val[2],table,0));
		b_val += b_mod;
	    }
	    a_val += a_mod;
	}
	s_val += s_mod;
    }

#endif

//    progress.show();
//    progress.set_text("finding surface..");
    for(int i = 0 ; i < state->gui.size.x.value() ; i++) {
// 	progress.update((double)i/(size.x-1));
	for(int j = 0 ; j < state->gui.size.y.value() ; j++) 
	    for(int k = 0 ; k < state->gui.size.z.value() ; k++) 
		props->set(i, j, k, SURFACE, 
			   !props->get(i, j, k, DIVERGES_1)
			   && divergentNeibhor(i, j, k));
    }
//    progress.hide();
}

void fractal::compute_projection() {

#define DIV 10

  //    printf("compute proj\n");


    if(!this || !state)
	return;

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();

    //    printf("mid compute state = %p\n", state);

    if(state->lw.view.perspective_on) {
      //	printf("persp on\n");
	glFrustum(-state->gui.size.x.value()/DIV,
 		  state->gui.size.x.value()/DIV,
 		  -state->gui.size.y.value()/DIV,
 		  state->gui.size.y.value()/DIV, 
 		  state->gui.size.z.value(), 
 		  -state->gui.size.z.value()*4);
    } else			// orthographic projection mode is currently 
				// broken :<
// 	glOrtho(-fract->state->gui.size.x.value(), 
// 		fract->state->gui.size.x.value(),
// 		-fract->state->gui.size.y.value(),
// 		fract->state->gui.size.y.value(), 
// 		fract->state->gui.size.z.value()*2,
// 		-fract->state->gui.size.z.value()*2);
	glOrtho(-DIV, DIV, -DIV, DIV, DIV, -DIV);
    
    //    printf("later\n");

    
    gluLookAt(0,0,6*state->gui.size.z.value(), // from position 
	      0,0,0,		// to position
	      0,1,0);		// up vector 


//     double zoom_val =  state->view.view ? state->view.zoom_val :
// 	GTK_ADJUSTMENT(state->view.view->zoom_adj->gtkobject)->value;
    double zoom_val = state->lw.view.get_zoom_val();
    
    //    double zoom_val = 1.0;
    
//    printf("in refresh zoom_val = %.2f\n", zoom_val);
    
    glScalef(zoom_val, zoom_val, zoom_val);
    glRotatef(state->lw.view.rot.gety(), 1.0, 0.0, 0.0);
    glRotatef(state->lw.view.rot.getx(), 0.0, 1.0, 0.0);

//    glScalef(state->view.view->,state->view.zoom_val,state->view.zoom_val);

    //    printf("rotating to %lf %lf\n", state->view.rot.getx(), state->view.rot.gety());


    glMatrixMode( GL_MODELVIEW );

    //    printf("exit compute proj\n");

}

void fractal::refresh_impl()	// default impl for refresh signal
{
//    clock_t start = clock();

//    printf("refresh_init\n");

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
//    printf("pre view\n");

    compute_projection();
    
//    printf("pre light\n");

    draw_light();
    
//    printf("post light\n");

    glMatrixMode(GL_MODELVIEW);	// draw fractal
    glLoadIdentity();
    glRotatef(-state->lw.model.rot.gety(), 1.0, 0.0, 0.0);
    glRotatef(-state->lw.model.rot.getx(), 0.0, 1.0, 0.0);

    glTranslatef(-((double)(state->gui.size.x.value()))/2,
		 -((double)(state->gui.size.y.value()))/2, 
		 -((double)(state->gui.size.z.value()))/2);

    glDrawElements(GL_QUADS, poly_count*4, GL_UNSIGNED_INT, indices);

//    printf("refresh_post drawElements\n");

    if(state->gui.bounding_box_visible)
	drawBoundingBox();

    if(state->gui.zoom_box_visible)
	drawZoomBox();

//    printf("refresh : pre swapBuffers\n");
//    printf("drawable = %p\n", drawable);
  
    glXSwapBuffers(dpy,drawable);

//    printf("refresh_exit\n");

//    clock_t finish = clock();

//     printf("refresh took %.2f seconds CPU time\n", 
// 	   (double)(finish - start)/CLOCKS_PER_SEC);
}

// draws model from surface representation (as boxes for now)

void fractal::computeSurface() {

//     hash->clear();		// first clear all revelant data structures

    if(first_pt) {
	delete first_pt;
	first_pt = last_pt = NULL;
    }

    for(int i = 0 ; i < state->gui.size.x.value() ; i++) 
	for(int j = 0 ; j < state->gui.size.y.value() ; j++)
	    for(int k = 0 ; k < state->gui.size.z.value() ; k++)
		position[i][j][k].point = NULL;
    
    if(first_poly) {
 	delete first_poly;
 	first_poly = NULL;
    }

    if(interleave) {
	delete [] interleave;
	interleave = NULL;
    }

    poly_count = 0;
    
//     progress.set_text("generating polygons"); // generate polygons
//     progress.show();
    for(int i = 0 ; i < state->gui.size.x.value() ; i++) {
//	progress.update((double)i/(state->gui.size.x.value()-1));
	for(int j = 0 ; j < state->gui.size.y.value() ; j++)
	    for(int k = 0 ; k < state->gui.size.z.value() ; k++) 
		if(props->get(i, j, k, SURFACE)) {
		    //	printf("^^%2f ", position[i][j][k].color);
		    drawCube(i, j, k, position[i][j][k].color);
		}
    }
//    progress.hide();

    if(!first_pt) {
	printf("no points\n");
//	assert(!interleave);

	interleave = NULL;
	return;
    }

    struct fcoord * p = first_pt;
    
    interleave = new float[(last_pt->index+1)*10];
    float * inter_ptr = interleave;

    p = first_pt;

    while(p) {
//	p->normalizeNormal();

	memcpy(inter_ptr, p, sizeof(float)*10);

	inter_ptr += 10;
	p = p->next;
    }

    glInterleavedArrays(GL_C4F_N3F_V3F, 0, interleave);

    indices = new unsigned int[poly_count*4];
    unsigned int *index_ptr = indices;

    for(fpoly * poly = first_poly ; poly ; poly = poly->next) {
	for(int i = 0 ; i < 4 ; i++)
	    index_ptr[i] = poly->vert[i];
	index_ptr += 4;
    }

    printf("%d points processed\n", last_pt->index+1);

    delete first_pt;
    first_pt = last_pt = NULL;

    delete first_poly;
    first_poly = NULL;
}

void fcoord::setColor(float c) {

//    printf("setting color %.2f\n", c);
    
    c /= 600;

    switch(1) {
    case 1:
	color[0] += /*(char)*/(c*c);
	color[1] += /*(char)*/c;
	color[2] += /*(char)*/((1-c)*(1-c));
       	break;

    case 2:
	color[0] += /*(char)*/c;
	color[1] += /*(char)*/(c*c);
	color[2] += /*(char)*/c;
	break;	

    case 3:
	color[0] += /*(char)*/c;
	color[1] += /*(char)*/c;
	color[2] += /*(char)*/c;
	break;

    case 4:
	color[0] += /*(char)*/(1-c);
	color[1] += /*(char)*/exp(c);
	color[2] += /*(char)*/c;
	break;

    case 5:
	color[0] += 0;
	color[1] += c/2;
	color[2] += 0.2;
	break;

    }

    //    printf("%.2f %d %d %d\n", c, color[0], color[1], color[2]);
}

int fractal::addPoint(int x_pos, int y_pos, int z_pos,
		      int x_norm, int y_norm, int z_norm, float color) {

    struct fcoord * new_pt;

    if(position[x_pos][y_pos][z_pos].point) { // add stuff
      
	fcoord * point = position[x_pos][y_pos][z_pos].point;

	point->norm[0] += x_norm;
	point->norm[1] += y_norm;
	point->norm[2] += z_norm;

	point->setColor(color);

	point->color[0] /= 2;
	point->color[1] /= 2;
	point->color[2] /= 2;
	
	return point->index;

    } else {

	new_pt = new fcoord;

	position[x_pos][y_pos][z_pos].point = new_pt;

	if(first_pt == NULL) {
	    first_pt = last_pt = new_pt;
	    new_pt->index = 0;
	    
	} else {

	    new_pt->index = last_pt->index + 1;

	    last_pt->next = new_pt;
	    last_pt = new_pt;
	    new_pt->next = NULL;
	}

	new_pt->pos[0] = x_pos;
	new_pt->pos[1] = y_pos;
	new_pt->pos[2] = z_pos;

	new_pt->norm[0] = x_norm;
	new_pt->norm[1] = y_norm;
	new_pt->norm[2] = z_norm;

	new_pt->setColor(color);

	return new_pt->index;
    }
}
    
fpoly::~fpoly() {
    if(next)
	delete next;
}

// adds a polygon to current image
void fractal::addPoly(int i1, int i2, int i3, int i4) {
    
    // first construct unique string representation of the coords
    char buf[256];
    sprintf(buf, "%d %d %d %d", i1, i2, i3, i4);

    if(/*hash->enter(buf)*/1) {	// avoid duplicates by hashing each as we go

	struct fpoly * new_poly = new fpoly;

	poly_count++;

	if(first_poly == NULL) 
	    first_poly = last_poly = new_poly;
	else {
	    last_poly->next = new_poly;
	    last_poly = new_poly;
	}
       
	new_poly->vert[0] = i1;
	new_poly->vert[1] = i2;
	new_poly->vert[2] = i3;
	new_poly->vert[3] = i4;
    
	new_poly->next = NULL;
    }
}
		      
void fractal::drawCube(int i, int j, int k, float color) { // for flat shading

    int i_plus = i+1,
	i_minus = i,
	j_plus = j+1,
	j_minus = j,
	k_plus = k+1,
	k_minus = k;

    /* first draw bottom square and vertical lines */
    

    if(i_plus == state->gui.size.x.value() || j_plus == state->gui.size.y.value() || k_plus == state->gui.size.z.value())
	return;

    addPoly(addPoint(i_minus,j_minus,k_minus,
		     -1, 0, 0,
		     color),
	    addPoint(i_minus, j_plus,k_minus,
		     -1, 0, 0,
		     color),
	    addPoint(i_minus, j_plus, k_plus,
		     -1, 0, 0,
		     color),
	    addPoint(i_minus,j_minus, k_plus,
		     -1, 0, 0,
		     color));

    addPoly(addPoint(i_plus,j_minus,k_minus,
		     1, 0, 0,
		     color),
	    addPoint(i_plus, j_plus,k_minus,
		     1, 0, 0,
		     color),
	    addPoint(i_plus, j_plus, k_plus,
		     1, 0, 0,
		     color),
	    addPoint(i_plus,j_minus, k_plus,
		     1, 0, 0,
		     color));

    addPoly(addPoint(i_plus,j_minus,k_minus,
		     0, -1, 0,
		     color),
	    addPoint(i_minus,j_minus,k_minus,
		     0, -1, 0,
		     color),
	    addPoint(i_minus,j_minus, k_plus,
		     0, -1, 0,
		     color),
	    addPoint(i_plus,j_minus, k_plus,
		     0, -1, 0,
		     color));

    addPoly(addPoint(i_plus,j_plus,k_minus,
		     0, 1, 0,
		     color),
	    addPoint(i_minus,j_plus,k_minus,
		     0, 1, 0,
		     color),
	    addPoint(i_minus,j_plus, k_plus,
		     0, 1, 0,
		     color),
	    addPoint(i_plus,j_plus, k_plus,
		     0, 1, 0,
		     color));

    addPoly(addPoint(i_plus,j_minus,k_minus,
		     0, 0, -1,
		     color),
	    addPoint(i_minus,j_minus,k_minus,
		     0, 0, -1,
		     color),
	    addPoint(i_minus, j_plus,k_minus,
		     0, 0, -1,
		     color),
	    addPoint(i_plus, j_plus,k_minus,
		     0, 0, -1,
		     color));

    addPoly(addPoint(i_plus,j_minus,k_plus,
		     0, 0, 1,
		     color),
	    addPoint(i_minus,j_minus,k_plus,
		     0, 0, 1,
		     color),
	    addPoint(i_minus, j_plus,k_plus,
		     0, 0, 1,
		     color),
	    addPoint(i_plus, j_plus,k_plus,
		     0, 0, 1,
		     color)); 
}

void fractal::recompute(bool sizechange) {
    identifySurface(sizechange);
    computeSurface();
}

void fractal::drawZoomBox() {

    double maxval_x = state->gui.maxval.x.value();
    double maxval_y = state->gui.maxval.y.value();
    double maxval_z = state->gui.maxval.z.value();

    double minval_x = state->gui.minval.x.value();
    double minval_y = state->gui.minval.y.value();
    double minval_z = state->gui.minval.z.value();

    double curr_min_x = state->gui.minval.x.gui_value();
    double curr_min_y = state->gui.minval.y.gui_value();
    double curr_min_z = state->gui.minval.z.gui_value();

    double curr_max_x = state->gui.maxval.x.gui_value();
    double curr_max_y = state->gui.maxval.y.gui_value();
    double curr_max_z = state->gui.maxval.z.gui_value();

    //    printf("curr_max_x = %.2f\n", curr_max_x);


//     curr_min_x *= state->gui.zoom.x.gui_value();
//     curr_min_y *= state->gui.zoom.y.gui_value();
//     curr_min_z *= state->gui.zoom.z.gui_value();
//     curr_max_x *= state->gui.zoom.x.gui_value();
//     curr_max_y *= state->gui.zoom.y.gui_value();
//     curr_max_z *= state->gui.zoom.z.gui_value();

    
    double new_min_x = (curr_min_x-minval_x)/(maxval_x-minval_x)*
	state->gui.size.x.value();
    double new_min_y = (curr_min_y-minval_y)/(maxval_y-minval_y)*
	state->gui.size.y.value();
    double new_min_z = (curr_min_z-minval_z)/(maxval_z-minval_z)*
	state->gui.size.z.value();
    
    double new_max_x = (curr_max_x-minval_x)/(maxval_x-minval_x)*
	state->gui.size.x.value();
    double new_max_y = (curr_max_y-minval_y)/(maxval_y-minval_y)*
	state->gui.size.y.value();
    double new_max_z = (curr_max_z-minval_z)/(maxval_z-minval_z)*
	state->gui.size.z.value();

//     new_min_x += state->gui.pan.x.gui_value();
//     new_min_y += state->gui.pan.y.gui_value();
//     new_min_z += state->gui.pan.z.gui_value();
//     new_max_x += state->gui.pan.x.gui_value();
//     new_max_y += state->gui.pan.y.gui_value();
//     new_max_z += state->gui.pan.z.gui_value();

    //    printf("new_max_x = %.2f\n", new_max_x);

    //    printf("check end\n");

    glBegin(GL_LINE_LOOP);
    
    glColor3f(1.0, 1.0, 1.0);
    
    glVertex3f(new_min_x, new_min_y, new_min_z);
    glVertex3f(new_min_x, new_max_y, new_min_z);
    glVertex3f(new_max_x, new_max_y, new_min_z);
    glVertex3f(new_max_x, new_min_y, new_min_z);

    glEnd();

    glBegin(GL_LINE_LOOP);

    glVertex3f(new_min_x, new_min_y, new_max_z);
    glVertex3f(new_min_x, new_max_y, new_max_z);
    glVertex3f(new_max_x, new_max_y, new_max_z);
    glVertex3f(new_max_x, new_min_y, new_max_z);

    glEnd();

    glBegin(GL_LINES);

    glVertex3f(new_min_x, new_min_y, new_max_z);
    glVertex3f(new_min_x, new_min_y, new_min_z);

    glVertex3f(new_min_x, new_max_y, new_max_z);
    glVertex3f(new_min_x, new_max_y, new_min_z);

    glVertex3f(new_max_x, new_min_y, new_max_z);
    glVertex3f(new_max_x, new_min_y, new_min_z);

    glVertex3f(new_max_x, new_max_y, new_max_z);
    glVertex3f(new_max_x, new_max_y, new_min_z);

    glEnd();

}

void fractal::drawBoundingBox() {
    
    float incr;

    glBegin(GL_LINES);

    glColor3f(1.0, 0.0, 0.0);

    for(float i = 0 ; i <= state->gui.bounding_res.value() ; i++) {
	
	incr = i/state->gui.bounding_res.value();

	glVertex3f(0, 0, incr*state->gui.size.z.value());
	glVertex3f(state->gui.size.x.value(), 0, 
		   incr*state->gui.size.z.value());

	glVertex3f(0, state->gui.size.y.value(),
		   incr*state->gui.size.z.value());
	glVertex3f(state->gui.size.x.value(), 
		   state->gui.size.y.value(), 
		   incr*state->gui.size.z.value());

	glVertex3f(state->gui.size.x.value(),
		   state->gui.size.y.value(),
		   incr*state->gui.size.z.value());
	glVertex3f(state->gui.size.x.value(),
		   0, incr*state->gui.size.z.value());

	glVertex3f(0, state->gui.size.y.value(), 
		   incr*state->gui.size.z.value());
	glVertex3f(0, 0, 
		   incr*state->gui.size.z.value());


	glVertex3f(incr*state->gui.size.x.value(), 0, 0);
	glVertex3f(incr*state->gui.size.x.value(), 0, 
		   state->gui.size.z.value());

	glVertex3f(incr*state->gui.size.x.value(), 
		   state->gui.size.y.value(), 0);
	glVertex3f(incr*state->gui.size.x.value(),
		   state->gui.size.y.value(),
		   state->gui.size.z.value());

	glVertex3f(incr*state->gui.size.x.value(),
		   state->gui.size.y.value(), 
		   state->gui.size.z.value());
	glVertex3f(incr*state->gui.size.x.value(), 0,
		   state->gui.size.z.value());

	glVertex3f(incr*state->gui.size.x.value(),
		   state->gui.size.y.value(), 0);
	glVertex3f(incr*state->gui.size.x.value(), 0, 0);


	glVertex3f(0, incr*state->gui.size.y.value(), 0);
	glVertex3f(0, incr*state->gui.size.y.value(),
		   state->gui.size.z.value());

	glVertex3f(state->gui.size.x.value(),
		   incr*state->gui.size.y.value(), 0);
	glVertex3f(state->gui.size.x.value(),
		   incr*state->gui.size.y.value(), 
		   state->gui.size.z.value());

	glVertex3f(0, incr*state->gui.size.y.value(),
		   state->gui.size.z.value());
	glVertex3f(state->gui.size.x.value(), 
		   incr*state->gui.size.y.value(),
		   state->gui.size.z.value());

	glVertex3f(0, incr*state->gui.size.y.value(), 0);
	glVertex3f(state->gui.size.x.value(), 
		   incr*state->gui.size.y.value(), 0);
    }

    glEnd();
}


void fractal::draw_light() {

    if(this && state->lw.light.on) {
  
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
 	glEnable(GL_LIGHTING);
 	glEnable(GL_LIGHT0);

//	radius = fract->size.x;
	state->lw.light.pos[Z] = 0;

   	glRotatef(-state->lw.light.rot.gety(), 1.0, 0.0, 0.0);
   	glRotatef(-state->lw.light.rot.getx(), 0.0, 1.0, 0.0);
	glTranslatef( 0, 0, state->gui.size.x.value()
		      *state->lw.light.getRadius());
	// 	float scale_val = state->lw.light.getRadius();
				// this one scales the light itself..
	// 	glScalef(scale_val, scale_val, scale_val);
				// next declare the light	
   	glMaterialfv(GL_FRONT, GL_AMBIENT, state->lw.light.mat_amb);
//   	glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_dif);
//  	glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
//   	glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
   	glMaterialfv(GL_FRONT, GL_EMISSION, state->lw.light.mat_emission);
 
 	glLightfv( GL_LIGHT0, GL_POSITION,  state->lw.light.pos );
 	glLightfv( GL_LIGHT0, GL_SPOT_DIRECTION, state->lw.light.at);

 	glLightfv( GL_LIGHT0, GL_AMBIENT,  state->lw.light.ambient );
 	glLightfv( GL_LIGHT0, GL_SPECULAR, state->lw.light.specular );
 	glLightfv( GL_LIGHT0, GL_DIFFUSE,  state->lw.light.diffuse );

	float unit = (float)(state->gui.size.x.value())/20.0;

	glBegin(GL_TRIANGLE_FAN);

	glColor3f( 0.0, 0.0, 1.0);
	glVertex3f(0,0,unit);
 	glColor3f( 1.0, 1.0, 1.0 );
	glVertex3f( unit, unit, -unit );
	glVertex3f(-unit, unit, -unit );
	glVertex3f(-unit,-unit, -unit );
	glVertex3f( unit,-unit, -unit );
	glVertex3f( unit, unit, -unit );
	
	glEnd();
//  	glBegin( GL_LINE_LOOP );
//  	/* first draw bottom square and vertical lines */
//  	glVertex3f( -unit,-unit,-unit );
//  	glVertex3f( -unit, unit,-unit );
//  	glVertex3f( -unit,-unit,-unit );
//  	glVertex3f(  unit,-unit,-unit );
//  	glVertex3f(  unit, unit,-unit );
//  	glVertex3f(  unit,-unit,-unit );
//  	glVertex3f(  unit,-unit, unit );
//  	glVertex3f(  unit, unit, unit );
//  	glVertex3f(  unit,-unit, unit );
//  	glVertex3f( -unit,-unit, unit );
//  	glVertex3f( -unit, unit, unit );
//  	glVertex3f( -unit,-unit, unit );
//  	glEnd();

//  	glBegin( GL_LINE_LOOP );
//  	/* now draw top square of the light source */
//  	glVertex3f( -unit, unit,-unit );
//  	glVertex3f(  unit, unit,-unit );
//  	glVertex3f(  unit, unit, unit );
//  	glVertex3f( -unit, unit, unit );
//  	glEnd();


 	glPopMatrix();  

     } else {
	glDisable(GL_LIGHTING);
     }
}
