/*
 *  fractal.h
 * 
 *  Fractal Generator
 * 
 *  Copyright (c) 1998 Brian Martin
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef FRACTAL_H
#define FRACTAL_H

#include <GL/glx.h>
#include <gtk--.h>

#ifdef USE_THREADS
#include <pthread.h>
#include <semaphore.h>
#include "queue.h"
#endif // USE_THREADS

#include "bitvect.h"
#include "hash.h"
#include "expr_tree.h"
#include "progress.h"
#include "light.h"
#include "glport.h"
#include "state.h"
#include "eq_table.h"
#include "fractal_state.h"

struct fcoord {

    float color[4];		// the order of these first three elements
    float norm[3];		// is important
    float pos[3];
    struct fcoord * next;
    int index;

    fcoord() {
	color[0] = 0;
	color[1] = 0;
	color[2] = 0;
	color[3] = 1.0;
	next = 0;
    }
    
    ~fcoord() {
	if(next)
	    delete next;
    }

    void normalizeNormal() {
	float sum = norm[0]+norm[1]+norm[2];
	norm[0] /= sum;
	norm[1] /= sum;
	norm[2] /= sum;
    }
    void setColor(float color);
};


#define DIVERGES_0 0
#define DIVERGES_1 1
#define DIVERGES_2 2
#define SURFACE    3

#define FRACT_RECOMPUTE_REQ 0
#define FRACT_REFRESH_REQ 1
#define FRACT_RECOMPUTE_SIZECHANGE_REQ 3

struct fractalPos {
    fcoord * point;
    float color;

    fractalPos(): point(0), color(0) {}
  
    void writeTo(FILE * file) { 

// 	fprintf(file, " %c ", prop); 
// 	if(prop & SURFACE_1)
// 	    fprintf(file, " %.5f ", color); 
    }

    void readFrom(FILE * file) { 

// 	fscanf(file, " %c ", &prop); 
// 	if(prop & SURFACE_1) 
// 	    fscanf(file, " %f ", &color); 
    }
};


struct fpoly {			// these are all squares
    
    int vert[4];		// vertex indices
    struct fpoly * next;
    fpoly(): next(0) {}
//    ~fpoly() { if(next) delete next; }
    ~fpoly();
};

class coordList {
    
private:
  
    class coordList * nextptr;

public:

    int a, b, c, d;

    coordList(int _a, int _b, int _c, int _d) :
	a(_a), b(_b), c(_c), d(_d), nextptr(0) {}
    ~coordList(){ if (nextptr != NULL) delete nextptr; }
    coordList * next() { return nextptr; }
    coordList *prepend(coordList * list) { list->nextptr = this; return list; }
    
};

class fractal : public virtual Gtk_Signal_Base {
   
private:

public:

    class fractal_State *state;

    bitvect * props;		// bitvector for surface and divergance info
				// this is used to segregate bit quantities
				// from word quantities

    fractalPos ***position;	// position properties (words)

    double max_val;		// max escape value
    double max_iter;		// max iterations for divergence
    
    double w_val[3];		// values for w point
    double s_mod, a_mod, b_mod;	// increments;

    // polygons and points for the rendered model

    struct fcoord * first_pt, * last_pt; // these lists are for point and
    struct fpoly * first_poly, * last_poly; // polygon creation
    int poly_count;
    float * interleave;		// interleaved array of packed vertex info
    unsigned int * indices;	// indexes into interleave specifying polygons
    
    Hash * hash;		// used for duplicate polygon removal

    class Fract_View *view;
    class Fract_Params *gui;

    Display *dpy;
    Window drawable;

    GLPort * glwin;

    Fract_Progress progress;

    Signal0 refresh;

    void free_position();

#ifdef USE_THREADS
    queue requests;
    pthread_t req_thread;

    int num_threads;
    int currthread;

    int get_thread_num();	// called by divergence threads
    
    eq_table ** table;

    int currblock;
    sem_t diverg_mutex;
    sem_t x_mutex;		// for mutex on X stuff
    
    void set_table(expr_ref expr) {
	for(int i = 0 ; i < num_threads ; i++)
	    delete table[i];
	table[0] = new eq_table(expr, max_iter, max_val);
	for(int i = 1 ; i < num_threads ; i++) 
	    table[i] = new eq_table(table[0]);
    }

    void change_num_threads(int newval) {
	if(newval > num_threads); // should probably do something here..
	
	num_threads = newval;
    }

    void request(short what) {
      // 	printf("queueing request %d\n", what);
 	requests.insert(what);
    }

//     void request(short what) {
// 	switch(what) {
// 	case FRACT_REFRESH_REQ:
// 	    refresh_impl();
// 	    break;

// 	case FRACT_RECOMPUTE_REQ:
// 	    recompute();
// 	    break;
// 	}
//     }


    void processes_requests();		// executed by req_thread
#else

    eq_table * table;

    void set_table(expr_ref expr) {
	delete table;
	table = new eq_table(expr, max_iter, max_val);
    }

    void request(short what) {
	switch(what) {
	case FRACT_REFRESH_REQ:
	    refresh_impl();
	    break;

	case FRACT_RECOMPUTE_SIZECHANGE_REQ:
	    recompute(true);
	    break;	    
	    
	case FRACT_RECOMPUTE_REQ:
	    recompute(false);
	    break;
	}
    }
#endif // USE_THREADS

    void clear() {
	first_pt = last_pt = 0;
	delete first_poly;
	first_poly = last_poly = 0;
    }

    void alloc();

//     void setFix(double _w) {
// 	w_fix = _w;
//     }

//     void setMinval(double x, double y, double z) {
// 	minval.x = x;
// 	minval.y = y;
// 	minval.z = z;
//     }

//     void setMaxval(double x, double y, double z) {
// 	maxval.x = x;
// 	maxval.y = y;
// 	maxval.z = z;
//     }

    fractal(class fractal_State *);
//    fractal(int, int, int);
				// finds surface based upon divergence info
    void identifySurface(bool sizechange); 
    void computeSurface();	// generates polygons from surface

    // determines if point has any divergent neibhors
    bool divergentNeibhor(int i, int j, int k); 

    // determines convergance (divergance) of point
    bool divergent(double, double, double, double, eq_table *, float*);	
#ifdef USE_THREADS
    void test_divergence();	// for threads
    int get_next_divergence_block(); 
#endif // USE_THREADS

    // lists neibhors of given point
    coordList * enumerateNeibhors(int a, int b, int c, int d); 

    void draw();		// draws model from polygons
    void drawBox(int i, int j, int k, float); // low level drawing primitives
    void drawCube(int i, int j, int k, float color);	
    int addPoint(int x_pos, int y_pos, int j_pos,
		 int x_norm, int y_norm, int j_norm,
		 float color);
    void addPoint(int x_pos, int y_pos, int j_pos, float color);
    void addPoly(int, int, int, int);

    void readFrom(FILE * file);	// these don't work right now..
    void writeTo(FILE * file);
    
				// use given function for fractal generation
    void set_expr(const char * expr); 
    void init();		// sets up default paramaters
    void gl_init();		// sets up gl
    void recompute(bool);	// calls {compute,identify}Surface
    void refresh_impl();	// redraws fractal
    void queue_refresh();	// callback for refresh signal
    void drawBoundingBox();	
    void drawZoomBox();
    void compute_projection();
    void draw_light();
};

#endif




