#include "octree.h"
#include "assert.h"

octree::octree(int _level, tree_params * _params,
	       double min_x, double min_y, double min_z,
	       double x_span, double y_span, double z_span) {

  //    printf("octree::octree - level = %d", _level);
    polys = 0;
    
    params = _params;

    level = _level;		// init data
    
    min.x = min_x;
    min.y = min_y;
    min.z = min_z;
    
    span.x = x_span;
    span.y = y_span;
    span.z = z_span;

    if (level < params->min_level)
 	alloc_children();
    else 
	for(int i = 0 ; i < 8 ; i++)
	    child[i] = NULL;

//     printf("octree::octree this = %p\n", this);
}

void octree::alloc_children() {
	child[0] = new octree(level+1, params,
			      min.x, 
			      min.y,
			      min.z,
			      span.x/2, span.y/2, span.z/2);
	child[1] = new octree(level+1, params,
			      min.x+span.x/2,
			      min.y,
			      min.z,
			      span.x/2, span.y/2, span.z/2);
	child[2] = new octree(level+1, params,
			      min.x,
			      min.y+span.y/2,
			      min.z,
			      span.x/2, span.y/2, span.z/2);
	child[3] = new octree(level+1, params,
			      min.x+span.x/2, 
			      min.y+span.y/2, 
			      min.z,
			      span.x/2, span.y/2, span.z/2);
	child[4] = new octree(level+1, params,
			      min.x, 
			      min.y,
			      min.z+span.z/2,
			      span.x/2, span.y/2, span.z/2);
	child[5] = new octree(level+1, params,
			      min.x+span.x/2,
			      min.y, 
			      min.z+span.z/2,
			      span.x/2, span.y/2, span.z/2);
	child[6] = new octree(level+1, params,
			      min.x, 
			      min.y+span.y/2,
			      min.z+span.z/2,
			      span.x/2, span.y/2, span.z/2);
	child[7] = new octree(level+1, params,
			      min.x+span.x/2,
			      min.y+span.y/2,
			      min.z+span.z/2,
			      span.x/2, span.y/2, span.z/2);

}

bool octree::not_void(int level) {

    bool not_void;		// cell is void if no polygons and no children

    if(level == 0) {
	not_void = child[0] || polys;
    } else if(child[0]) {	// return void only if all children are void
	not_void = (child[0]->not_void(level-1) ||
		    child[1]->not_void(level-1) ||
		    child[2]->not_void(level-1) ||
		    child[3]->not_void(level-1) ||
		    child[4]->not_void(level-1) ||
		    child[5]->not_void(level-1) ||
		    child[6]->not_void(level-1) ||
		    child[7]->not_void(level-1));
    } else {			// level > 1 && !children
	not_void = true;//polys != NULL;
    }
    return not_void;    
}

void octree::subdivide(int weight, double x_min, double y_min, double z_min,
		       double x_max, double y_max, double z_max) {

    if(((x_min > min.x && x_min < min.x + span.x) || // match cell
	(min.x > x_min && min.x < x_max)) && 
       ((y_min > min.y && y_min < min.y + span.y) ||
	(min.y > y_min && min.y < y_max)) && 
       ((z_min > min.z && z_min < min.z + span.z) ||
	(min.z > z_min && min.z < z_max)) &&
       not_void(params->fract->state->gui.tree_ignore_depth.gui_value()))
       
				// don't exceed maximum depth
	if(level < params->fract->state->gui.tree_max_depth.gui_value() - 1) {
	    if(child[0]) {
		for(int i = 0 ; i < 8 ; i++)
		    child[i]->subdivide(weight, 
					x_min, y_min, z_min, 
					x_max, y_max, z_max);

	    } else if(weight == 0) { // no more subdivision
	      if(polys)
		  remove_polys();

	      generate_polygons();

	    } else {		// no children, but should subdivide further
		if(polys)	// remove any existing polygons
		    remove_polys();

		alloc_children();

		for(int i = 0 ; i < 8 ; i++)
		    child[i]->subdivide(weight - 1,
					x_min, y_min, z_min,
					x_max, y_max, z_max);
	    }
	} else if(!child[0] && polys == NULL)
	    generate_polygons();
}

inline void octree::remove_polys() {
    List<fpoly*> * pl = polys;
    List<fpoly*> * old_pl;
    int count = 0;
    while(pl) {
	if(pl->data->next)
	    pl->data->next->prev = pl->data->prev;
		
	if(pl->data->prev)
	    pl->data->prev->next = pl->data->next;
	else
	    params->first_poly = pl->data->next;
		
	pl->data->clear_points();
		
	delete pl->data; // remove poly from list
	old_pl = pl;
	pl = pl->next;
	delete old_pl;
    }
    polys = NULL;
}

int octree::max_depth() {
    
    int levels[8];

    if(child[0]) {
	for(int i = 0 ; i < 8 ; i++)
	    levels[i] = child[i]->max_depth();
	int max = levels[0];
	for(int i = 1 ; i < 8 ; i++)
	    if(levels[i] > max) max = levels[i];
	return max;
    } else
	return level;
}

int octree::min_depth() {
    
    int levels[8];

    if(child[0]) {
	for(int i = 0 ; i < 8 ; i++)
	    levels[i] = child[i]->min_depth();
	int min = levels[0];
	for(int i = 1 ; i < 8 ; i++)
	    if(levels[i] < min) min = levels[i];
	return min;
    } else
	return level;
}

int tree_params::poly_count() {
    int count = 0;
    for(fpoly * f = first_poly ; f ; f = f->next) count++;
    return count;
}

int octree::count() {
    
    int count = 0;

    for(int i = 0 ; i < 8 ; i++) if(child[i]) count += child[i]->count();
    return(count+1);
}

void octree::print_polys() {

    printf("%p ", polys);
    if(child[0])
	for(int i = 0 ; i < 8 ; i ++)
	    child[i]->print_polys();

}

octree::~octree() {
    for(int i = 0 ; i < 8 ; i++)
	if(child[i])
 	    delete child[i];
}

void octree::generate_polygons() {

    bool diverg[9];


    if(child[0]) {		// assuming none or all children allocated
	for(int i = 0 ; i < 8 ; i++)
	    child[i]->generate_polygons();
    } else {


	diverg[0] = params->fract->divergent(min.x, 
					     min.y,
					     min.z,
					     params->fract->w_val[1], 
					     params->table,
					     0); 
	diverg[1] = params->fract->divergent(min.x+span.x,
					     min.y,
					     min.z,
					     params->fract->w_val[1], 
					     params->table, 
					     0);
	diverg[2] = params->fract->divergent(min.x,
					     min.y+span.y,
					     min.z,
					     params->fract->w_val[1],
					     params->table, 
					     0);
	diverg[3] = params->fract->divergent(min.x+span.x,
					     min.y+span.y,
					     min.z,
					     params->fract->w_val[1],
					     params->table, 
					     0);

	diverg[4] = params->fract->divergent(min.x,
					     min.y,
					     min.z+span.z,
					     params->fract->w_val[1],
					     params->table,
					     0);
	diverg[5] = params->fract->divergent(min.x+span.x,
					     min.y,
					     min.z+span.z,
					     params->fract->w_val[1],
					     params->table,
					     0);
	diverg[6] = params->fract->divergent(min.x,
					     min.y+span.y,
					     min.z+span.z, 
					     params->fract->w_val[1], 
					     params->table, 
					     0);
	diverg[7] = params->fract->divergent(min.x+span.x,
					     min.y+span.y,
					     min.z+span.z, 
					     params->fract->w_val[1],
					     params->table,
					     0);
	diverg[8] = params->fract->divergent(min.x+span.x/2,
					     min.y+span.y/2,
					     min.z+span.z/2,
					     params->fract->w_val[1],
					     params->table, 
					     0);

	coord coords[9];
	int corner[4];

	coords[0].x = min.x;
	coords[0].y = min.y;
	coords[0].z = min.z;

	coords[1].x = min.x+span.x;
	coords[1].y = min.y;
	coords[1].z = min.z;

	coords[2].x = min.x;
	coords[2].y = min.y+span.y;
	coords[2].z = min.z;

	coords[3].x = min.x+span.x;
	coords[3].y = min.y+span.y;
	coords[3].z = min.z;

	coords[4].x = min.x;
	coords[4].y = min.y;
	coords[4].z = min.z+span.z;

	coords[5].x = min.x+span.x;
	coords[5].y = min.y;
	coords[5].z = min.z+span.z;

	coords[6].x = min.x;
	coords[6].y = min.y+span.y;
	coords[6].z = min.z+span.z;

	coords[7].x = min.x+span.x;
	coords[7].y = min.y+span.y;
	coords[7].z = min.z+span.z;

	coords[8].x = min.x+span.x/2;
	coords[8].y = min.y+span.y/2;
	coords[8].z = min.z+span.z/2;	

	corner[0] = 8;

	corner[1] = 1;	// lesser plane in x-y
	corner[2] = 2;
	corner[3] = 0;

// 	printf("make 1\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 2;
	corner[2] = 1;
	corner[3] = 3;

// 	printf("make 2\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 5;	// greater plane in x-y
	corner[2] = 4;
	corner[3] = 6;

// 	printf("make 3\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 6;
	corner[2] = 7;
	corner[3] = 5;

// 	printf("make 4\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 4;	// lesser plane in y-z
	corner[2] = 0;
	corner[3] = 2;

// 	printf("make 5\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 4;
	corner[2] = 2;
	corner[3] = 6;

// 	printf("make 6\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 1;	// greater plane in y-z
	corner[2] = 5;
	corner[3] = 3;

// 	printf("make 7\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 3;
	corner[2] = 5;
	corner[3] = 7;

// 	printf("make 8\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 1;	// lesser plane in x-z
	corner[2] = 0;
	corner[3] = 4;

// 	printf("make 9\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 1;
	corner[2] = 4;
	corner[3] = 5;

// 	printf("make 10\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 3;	// greater plane in x-z
	corner[2] = 6;
	corner[3] = 2;

// 	printf("make 11\n");
	make_polygons(corner,coords,diverg);

	corner[1] = 3;
	corner[2] = 7;
	corner[3] = 6;

// 	printf("make 12\n");
	make_polygons(corner,coords,diverg);

    }
}

int z_which = 0;

fcoord * octree::add_point(double x, double y, double z, float c) {

    double buf[sizeof(double)*3];

//     printf("add_point called with %.5f %.5f %.5f\n", x, y, z);

    HASH_TO_STRING_3(x,y,z,buf);

//     printf("adding point %.10f %.10f %.10f ...", x, y, z);
    Hash_node<fcoord *> * n = params->points->enter((char*)buf, true);

    if(!n->data) {
// 	printf("not found\n");
        n->data = new fcoord;

 	n->data->pos[0] = x;
 	n->data->pos[1] = y;
 	n->data->pos[2] = z;
// 	polys = NULL;
// 	memcpy(n->data->pos, buf, sizeof(double)*3);
				// color now computed from normal
//   	n->data->color[0] += (c*c);
//   	n->data->color[1] += c;
//   	n->data->color[2] += (1-c)*(1-c);
    }//  else {
//   	printf("cached\n");
    // }

    return n->data;
}

void octree::add_poly(coord x, float c1,
		      coord y, float c2,
		      coord z, float c3) {

    float v1[3];
    float v2[3];

    fcoord *  p1,* p2, * p3;
    
//     p1 = p2 = p3 = NULL;

    p1 = add_point(x.x, x.y, x.z, c1);
    p2 = add_point(y.x, y.y, y.z, c2);
    p3 = add_point(z.x, z.y, z.z, c3);


    fpoly * poly = new fpoly;

    poly->params = params;

				// compute polygon normal
    v1[0] = y.x - z.x;
    v1[1] = y.y - z.y;
    v1[2] = y.z - z.z;
    v2[0] = x.x - z.x;
    v2[1] = x.y - z.y;
    v2[2] = x.z - z.z;
    poly->norm[0] =   v1[1] * v2[2]  -  v1[2] * v2[1];
    poly->norm[1] = - v1[0] * v2[2]  +  v1[2] * v2[0];
    poly->norm[2] =   v1[0] * v2[1]  -  v1[1] * v2[0];
    
#if 0
    bool diverg = params->fract->divergent(x.x + 10*poly->norm[0], 
					   x.y + 10*poly->norm[1],
					   x.z + 10*poly->norm[2],
					   params->fract->w_val[1],
					   params->table,
					   0);

    printf("diverg = %d\n", diverg);
    if(diverg == false) {
	exit(-1);
    } else {

    }
#endif

//     double sum = poly->norm[0] + poly->norm[1] + poly->norm[2];
//     poly->norm[0] /= sum;
//     poly->norm[1] /= sum;
//     poly->norm[2] /= sum;

//     printf("poly %p:\n", poly);
//     printf("points %.5f %.5f %.5f\n", x.x, x.y, x.z);
//     printf("       %.5f %.5f %.5f\n", y.x, y.y, y.z);
//     printf("       %.5f %.5f %.5f\n", z.x, z.y, z.z);
//     printf("normal %.5f %.5f %.5f\n", poly->norm[0], poly->norm[1],
// 	   poly->norm[2]);
    
    poly->vert[0] = p1;
    poly->vert[1] = p2;
    poly->vert[2] = p3;

    DList<fpoly*> *l = new DList<fpoly*>(poly, p1->polys, 0);
    if(p1->polys && p1->polys->next)
	p1->polys->next->prev = l;
    p1->polys = l;
    
    l = new DList<fpoly*>(poly, p2->polys, 0);
    if(p2->polys && p2->polys->next)
	p2->polys->next->prev = l;
    p2->polys = l;
    
    l = new DList<fpoly*>(poly, p3->polys, 0);
    if(p3->polys && p3->polys->next)
	p3->polys->next->prev = l;
    p3->polys = l;
    
    poly->next = params->first_poly;
    if(params->first_poly) 
	params->first_poly->prev = poly;
    params->first_poly = poly;

    List<fpoly*> * pl = new List<fpoly*>(poly, polys);
    polys = pl;

// #define f_which(a) \
//     a[0] ? 'r' : (a[1] ? 'g' : (a[2] ? 'b' : 'x'))

//    printf("poly had indices %d %d %d\n", p1->index, p2->index, p3->index);
//    printf("poly had color %c%c%c\n", f_which(p1->color), f_which(p2->color), f_which(p3->color));
}

coord octree::search_for_point(coord a, coord b, int level) {

    char buf_a[sizeof(double)*4];
    char buf_b[sizeof(double)*4];
    char buf_n[sizeof(double)*4];

    coord new_point;
    new_point.x = (a.x+b.x)/2;
    new_point.y = (a.y+b.y)/2;
    new_point.z = (a.z+b.z)/2;

    if(level) {

	bool diverg = params->fract->divergent(new_point.x, 
					       new_point.y,
					       new_point.z,
					       params->fract->w_val[1],
					       params->table,
					       0);
    
	HASH_TO_STRING_4(a.x,a.y,a.z,params->fract->w_val[1],buf_a);
	HASH_TO_STRING_4(b.x,b.y,b.z,params->fract->w_val[1],buf_b);
	HASH_TO_STRING_4(new_point.x,new_point.y,new_point.z,
			 params->fract->w_val[1],buf_n);

	bool diverg_a, diverg_b, diverg_n;

	params->fract->diverg_cache->lookup(buf_a, diverg_a);
	params->fract->diverg_cache->lookup(buf_b, diverg_b);
	params->fract->diverg_cache->lookup(buf_n, diverg_n);
    
	if(diverg_a == diverg_n) {
	    return search_for_point(new_point, b, level - 1);
	} else if(diverg_n == diverg_b) {
	    return search_for_point(new_point, a, level - 1);
	} else if(diverg_a == diverg_b) {
	    fprintf(stderr, "points are same!!!\n");
	    exit(-1);
	}
    } else {
	return new_point;
    }
}

#if 0
#define octree_between(a, b) \
     (coords[corner[a]].x+coords[corner[b]].x)/2,\
     (coords[corner[a]].y+coords[corner[b]].y)/2,\
     (coords[corner[a]].z+coords[corner[b]].z)/2,\
     (color[corner[a]]+color[corner[b]])/2
#else
#define octree_between(a, b) \
     search_for_point(coords[corner[a]], coords[corner[b]], params->fract->state->gui.tree_bisect_iterations.value()), 1.0
#endif

#define swapp(a,b) b,a

void octree::make_polygons(int corner[4], coord coords[9], bool diverg[9]) {

//     printf("case %d%d%d%d\n", diverg[corner[0]], 
// 	   diverg[corner[1]], diverg[corner[2]], diverg[corner[3]]);
    
    if(diverg[corner[0]]) {	// binary search for correct case
	if(diverg[corner[1]]) {
	    if(diverg[corner[2]]) {
		if(diverg[corner[3]]) {	// case 1111
		    
		} else {	// case 1110 maybe?
// 		    printf("case 1110\n");
		    add_poly(octree_between(0,3), 
			     octree_between(2,3),
			     octree_between(1,3));
		}
	    } else {
		if(diverg[corner[3]]) {	// case 1101
// 		    printf("case 1101\n");
		    add_poly(octree_between(0,2),
			     octree_between(1,2),
			     octree_between(3,2));
		} else {	// case 1100 - ?
// 		    printf("case 1100\n");
		    add_poly(octree_between(1,3),
			     octree_between(0,2),
			     octree_between(1,2));

		    add_poly(octree_between(0,2),
			     octree_between(1,3),
			     octree_between(0,3));
		}
	    }
	} else {
	    if(diverg[corner[2]]) {
		if(diverg[corner[3]]) {	// case 1011
// 		    printf("case 1011\n");
		    add_poly(octree_between(0,1),// this one is right
			     octree_between(3,1),
			     octree_between(2,1));
		} else {	// case 1010 -
// 		    printf("case 1010\n");
		    add_poly(octree_between(0,1),
			     octree_between(0,3),
			     octree_between(2,1));

		    add_poly(octree_between(0,3),
			     octree_between(2,3),
			     octree_between(1,2));
		}
	    } else {
		if(diverg[corner[3]]) {	// case 1001 -
// 		    printf("case 1001\n");
		    add_poly(octree_between(2,0),
			     octree_between(0,1),
			     octree_between(1,3));

		    add_poly(octree_between(0,2),
			     octree_between(1,3),
			     octree_between(3,2));
		} else {	// case 1000
// 		    printf("case 1000\n");
		    add_poly(octree_between(2,0),// indifferent
			     octree_between(1,0),
			     octree_between(3,0));
		}
	    }
	}
    } else {
	if(diverg[corner[1]]) {
	    if(diverg[corner[2]]) {
		if(diverg[corner[3]]) {	// case 0111
// 		    printf("case 0111\n");
		    add_poly(octree_between(2,0),// indifferent
			     swapp(octree_between(1,0),
				   octree_between(3,0)));
		} else {	// case 0110 -
// 		    printf("case 0110\n");
		    add_poly(octree_between(2,0),
			     octree_between(2,3),
			     octree_between(1,0));

		    add_poly(octree_between(2,3),
			     octree_between(1,3),
			     octree_between(1,0));
		}
	    } else {
		if(diverg[corner[3]]) {	// case 0101 ?
// 		    printf("case 0101\n");
		    add_poly(octree_between(2,1),
			     octree_between(0,3),
			     octree_between(0,1));

		    add_poly(octree_between(2,3),
			     octree_between(0,3),
			     octree_between(1,2));
		} else {	// case 0100
// 		    printf("case 0100\n");
		    add_poly(octree_between(0,1),// indifferent
			     octree_between(2,1),
			     octree_between(3,1));
		}
	    }
	} else {
	    if(diverg[corner[2]]) {
		if(diverg[corner[3]]) {	// case 0011 ?
// 		    printf("case 0011\n");
		    add_poly(octree_between(1,2),
			     octree_between(0,2),
			     octree_between(1,3));

		    add_poly(octree_between(1,3),
			     octree_between(0,2),
			     octree_between(0,3));

		} else {	// case 0010
// 		    printf("case 0010\n");
		    add_poly(octree_between(1,2),// right
			     octree_between(0,2),
			     octree_between(3,2));
		}
	    } else {
		if(diverg[corner[3]]) {	// case 0001
// 		    printf("case 0001\n");
		    add_poly(octree_between(1,3),
			     octree_between(2,3), // indifferent
			     octree_between(0,3));
		} else {	// case 0000
		}
	    }
	}
    }
}

tree_params::tree_params() {
    points = NULL;
    fract = NULL;
    table = NULL;
    first_poly = NULL;    
}

tree_params::~tree_params() {	
    for(fpoly * f = first_poly ; f ; ) {
	fpoly * fuck = f;
	f = f->next;
	delete fuck;
    }
    if(points)
 	delete points;
    if(table) {
	printf("deleting table\n");
 	delete table;
    }
}
